# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._exceptions import Krb5Error
from krb5._keyblock import init_keyblock

from krb5._context cimport Context
from krb5._keyblock cimport KeyBlock
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    krb5_error_code krb5_c_string_to_key(
        krb5_context context,
        krb5_enctype enctype,
        const krb5_data *string,
        const krb5_data *salt,
        krb5_keyblock *key,
    ) nogil

    krb5_error_code krb5_c_string_to_key_with_params(
        krb5_context context,
        krb5_enctype enctype,
        const krb5_data *string,
        const krb5_data *salt,
        const krb5_data *params,
        krb5_keyblock *key,
    ) nogil


def c_string_to_key(
    Context context not None,
    krb5_enctype enctype,
    const unsigned char[:] string,
    const unsigned char[:] salt,
    const unsigned char[:] s2kparams = None,
) -> KeyBlock:
    cdef krb5_error_code err = 0
    cdef size_t length = 0

    cdef KeyBlock kb = init_keyblock(context, enctype, None)

    cdef krb5_data string_raw
    if len(string) == 0:
        pykrb5_set_krb5_data(&string_raw, 0, "")
    else:
        pykrb5_set_krb5_data(&string_raw, len(string), <char *>&string[0])

    cdef krb5_data salt_raw
    if len(salt) == 0:
        pykrb5_set_krb5_data(&salt_raw, 0, "")
    else:
        pykrb5_set_krb5_data(&salt_raw, len(salt), <char *>&salt[0])

    cdef krb5_data s2kparams_raw
    if s2kparams is None:
        err = krb5_c_string_to_key(context.raw, enctype, &string_raw, &salt_raw, kb.raw)
    else:
        if len(s2kparams) == 0:
           pykrb5_set_krb5_data(&s2kparams_raw, 0, "")
        else:
           pykrb5_set_krb5_data(&s2kparams_raw, len(s2kparams), <char *>&s2kparams[0])

        err = krb5_c_string_to_key_with_params(context.raw, enctype, &string_raw, &salt_raw, &s2kparams_raw, kb.raw)

    if err:
        raise Krb5Error(context, err)

    return kb



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import krb5


def test_cccol_iter() -> None:
    ctx = krb5.init_context()
    ccache = krb5.cc_new_unique(ctx, b"MEMORY")
    krb5.cc_initialize(ctx, ccache, krb5.parse_name_flags(ctx, b"username@DOMAIN.COM"))
    assert ccache.principal is not None
    krb5.cc_set_default_name(ctx, (ccache.cache_type or b"") + b":" + (ccache.name or b""))

    # MIT and Heimdal differ in the amounts returns, just make sure at least the MEMORY one is there
    actual = list(krb5.cccol_iter(ctx))
    assert len(actual) > 0
    for cache in actual:
        assert isinstance(cache, krb5.CCache)

    mem_ccache = next(iter([c for c in actual if c.cache_type == b"MEMORY"]))
    assert isinstance(mem_ccache, krb5.CCache)
    assert mem_ccache.cache_type == b"MEMORY"
    assert mem_ccache.name == ccache.name
    assert mem_ccache.principal is not None
    assert mem_ccache.principal.name == ccache.principal.name



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._krb5_types cimport *
from krb5._kt cimport KeyTab


cdef extern from "python_krb5.h":
    # Added in MIT 1.11
    krb5_error_code krb5_kt_have_content(
        krb5_context context,
        krb5_keytab keytab,
    ) nogil


def kt_have_content(
    Context context not None,
    KeyTab keytab not None,
) -> bool:
    cdef krb5_error_code err = 0

    err = krb5_kt_have_content(context.raw, keytab.raw)

    return err == 0



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._ccache import CCache
from krb5._context import Context

def cccol_iter(
    context: Context,
) -> typing.Iterator[CCache]:
    """Iterate over credential caches.

    Iterates over all known credential caches independent of type.

    Args:
        context: Krb5 context.

    Returns:
        Iterator[CCache]: An iterator of credential caches.
    """



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._context import Context
from krb5._keyblock import KeyBlock

def c_string_to_key(
    context: Context,
    enctype: int,
    string: bytes,
    salt: bytes,
    s2kparams: typing.Optional[bytes] = None,
) -> KeyBlock:
    """Convert a password string to a key.

    Convert a password string and a salt value plus optional S2K parameters
    to a keyblock with a certain encryption type. For getting the salt value,
    encryption type and the S2K parameters the method :meth:`get_etype_info`
    can be used.

    Args:
        context: Krb5 context.
        enctype: The encryption type to be used.
        string: The password string.
        salt: The salt string.
        s2kparams: The S2K parameters.

    Returns:
        KeyBlock: The keyblock.
    """



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._exceptions import Krb5Error

from krb5._ccache cimport CCache
from krb5._context cimport Context
from krb5._krb5_types cimport *
from krb5._principal cimport Principal


cdef extern from "python_krb5.h":
    # Added in MIT krb5 1.10
    krb5_error_code krb5_cc_cache_match(
        krb5_context context,
        krb5_principal client,
        krb5_ccache *cache_out,
    ) nogil



def cc_cache_match(
    Context context not None,
    Principal principal not None,
) -> CCache:
    ccache = CCache(context)
    cdef krb5_error_code err = 0

    err = krb5_cc_cache_match(context.raw, principal.raw, &ccache.raw)
    if err:
        raise Krb5Error(context, err)

    return ccache



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._context import Context
from krb5._keyblock import KeyBlock
from krb5._principal import Principal

class KeyTab:
    """Kerberos KeyTab object.

    This class represents a Kerberos key table.

    Args:
        context: Krb5 context.
    """

    def __iter__(self) -> typing.Iterator["KeyTabEntry"]:
        """Iterate through keytab entries.

        Enumerates through all the entries in a keytab. Will fail if the keytab
        being iterated does not exist. It may not be possible to add/remove
        entries on a keytab while it is being enumerated.
        """
    @property
    def addr(self) -> typing.Optional[int]:
        """The raw krb5_keytab pointer address of this credential cache."""
    @property
    def name(self) -> typing.Optional[bytes]:
        """The name/residual of the keytab."""
    @property
    def kt_type(self) -> typing.Optional[bytes]:
        """The type of the keytab."""

class KeyTabEntry:
    """Kerberos KeyTabEntry object.

    This class represents a key table entry.

    Note:
        When using a key tab entry from enumerating a :class:`KeyTab`, the
        key and principal value is tied to the lifetime of the entry. Attempting
        to use either of these after the entry is out of scope and has been
        freed will crash the process. The principal can be copied with
        `copy(entry.principal)` to ensure it outlives the entry context.
    """

    @property
    def key(self) -> KeyBlock:
        """The keytab key data block."""
    @property
    def kvno(self) -> int:
        """The key version number associated with the keytab entry."""
    @property
    def principal(self) -> Principal:
        """The principal associated with the keytab entry."""
    @property
    def timestamp(self) -> int:
        """The time creation entry of the keytab entry."""

def kt_add_entry(
    context: Context,
    keytab: KeyTab,
    principal: Principal,
    kvno: int,
    timestamp: int,
    keyblock: KeyBlock,
) -> None:
    """Add new keytab entry.

    Adds a new entry to a key table.

    Args:
        context: Krb5 context.
        keytab: Key table to add the entry to.
        principal: The principal to add in the keytab.
        kvno: The key version number to add to the keytab.
        timestamp: The seconds since EPOCH when the entry was added.
        keyblock: The key and encryption type of the entry to add.
    """

def kt_default(
    context: Context,
) -> KeyTab:
    """Resolve the default key table.

    Get a handle to the default keytab.

    Args:
        context: Krb5 context.

    Returns:
        KeyTab: The default keytab.
    """

def kt_default_name(
    context: Context,
) -> bytes:
    """Get the default key table name.

    Gets the name of the default key table for the context specified.

    Args:
        context: Krb5 context.

    Returns:
        bytes: The default key table name.
    """

def kt_get_entry(
    context: Context,
    keytab: KeyTab,
    principal: Principal,
    kvno: int = 0,
    enctype: int = 0,
) -> KeyTabEntry:
    """Get an entry from a key table.

    Retrieve an entry from a key table which matches the keytab, principal,
    kvno, and enctype. If kvno is 0, retrieve the highest-numbered kvno
    matching the other fields. If enctype is 0, match any enctype.

    Args:
        context: Krb5 context.
        keytab: The keytab to search.
        principal: The principal to match.
        kvno: The kvno to match in the keytab or 0 to match the highest kvno.
        enctype: The encryption type to get in the keytab or 0 to match any.

    Returns:
        KeyTabEntry: The entry found in the keytab.
    """

def kt_get_name(
    context: Context,
    keytab: KeyTab,
) -> bytes:
    """Get a key table name.

    Get the name of the specified key table. See :meth:`kt_get_type()` to get
    the type of a keytab.

    Note:
        MIT Kerberos returns the full name whereas Heimdal just returns the
        name/residual. See :meth:`krb5_get_full_name` for the equivalent in
        Heimdal.

    Args:
        context: Krb5 context.
        keytab: The keytab to query.

    Returns:
        bytes: The name of the keytab.
    """

def kt_get_type(
    context: Context,
    keytab: KeyTab,
) -> bytes:
    """Get a key table type.

    Get the type of the specified key table. See :meth:`ky_get_name()` to get
    the name of a keytab.

    Args:
        context: Krb5 context.
        keytab: The keytab to query.

    Returns:
        bytes: The type of the keytab.
    """

def kt_read_service_key(
    context: Context,
    name: typing.Optional[bytes],
    principal: Principal,
    kvno: int = 0,
    enctype: int = 0,
) -> KeyBlock:
    """Retrieve a service key from a key table.

    Args:
        context: Krb5 context.
        name: The keytable to open with :meth:`kt_resolve`, uses
            :meth:`kt_default` if None or an empty byte string.
        principal: The service principal to select in the keytab.
        kvno: The key version number or 0 for the highest available.
        enctype: The encryption type or 0 for any type.

    Returns:
        KeyBlock: The key associated with the keytab entry.
    """

def kt_remove_entry(
    context: Context,
    keytab: KeyTab,
    entry: KeyTabEntry,
) -> None:
    """Remove keytab entry.

    Removes an entry from a key table.

    Args:
        context: Krb5 context.
        keytab: Key table to remove the entry from.
        entry: The entry to be removed.
    """

def kt_resolve(
    context: Context,
    name: bytes,
) -> KeyTab:
    """Get a handle for a key table.

    Resolve the key table name and open a handle. The name must be of the from
    ``type:residual`` where type must be known to the library. If no type is
    specified then ``FILE`` is used as a default. The ``residual`` value is
    dependent on the type specified.

    Args:
        context: Krb5 context.
        name: The name of the keytab in the form ``type:residual``.

    Returns:
        KeyTab: The opened keytab.
    """



from typing import Any, Dict, List

class K5Realm:
    def __new__(cls, *args: Any, **kwargs: Any) -> K5Realm: ...
    tmpdir: str
    is_existing: bool
    realm: str
    portbase: int
    user_princ: str
    admin_princ: str
    host_princ: str
    nfs_princ: str
    krbtgt_princ: str
    keytab: str
    client_keytab: str
    ccache: str
    kadmin_ccache: str
    env: Dict[str, str]
    def __init__(
        self,
        realm: str = ...,
        portbase: int = ...,
        krb5_conf: Dict[str, str] | None = ...,
        kdc_conf: Dict[str, str] | None = ...,
        create_kdb: bool = ...,
        krbtgt_keysalt: str | None = ...,
        create_user: bool = ...,
        get_creds: bool = ...,
        create_host: bool = ...,
        start_kdc: bool = ...,
        start_kadmind: bool = ...,
        existing: str | None = ...,
        **paths: str,
    ) -> None: ...
    @property
    def provider(self) -> str: ...
    def create_kdb(self) -> None: ...
    def addprinc(
        self,
        princname: str,
        password: str | None = ...,
    ) -> None: ...
    def change_password(
        self,
        principal: str,
        password: str | None = ...,
        keysalt: str | None = ...,
    ) -> None: ...
    def extract_keytab(
        self,
        princname: str,
        keytab: str,
    ) -> None: ...
    def kinit(
        self,
        princname: str,
        password: str | None = ...,
        flags: List[str] | None = ...,
        verbose: bool = ...,
        **keywords: Any,
    ) -> str: ...
    def klist(
        self,
        ccache: str | None = ...,
        **keywords: Any,
    ) -> str: ...
    def klist_keytab(
        self,
        keytab: str | None = ...,
        **keywords: Any,
    ) -> str: ...
    def prep_kadmin(
        self,
        princname: str | None = ...,
        pw: str | None = ...,
        flags: List[str] | None = ...,
    ) -> str: ...
    def run_kadmin(
        self,
        query: str | List[str],
        **keywords: Any,
    ) -> str: ...
    def run_kadminl(
        self,
        query: str | List[str],
        **keywords: Any,
    ) -> Any: ...
    def start_kdc(
        self,
        args: List[str] | None = ...,
        env: Dict[str, str] | None = ...,
    ) -> None: ...
    def start_kadmind(
        self,
        env: Dict[str, str] | None = ...,
    ) -> None: ...
    @property
    def hostname(self) -> str: ...
    def run(
        self,
        args: List[str],
        env: Dict[str, str] | None = ...,
        input: str | None = ...,
        expected_code: int = ...,
    ) -> str: ...
    def __del__(self) -> None: ...
    def kprop_port(self) -> int: ...
    def server_port(self) -> int: ...
    def stop_kdc(self) -> None: ...
    def stop_kadmind(self) -> None: ...
    def stop(self) -> None: ...
    def password(
        self,
        name: str,
    ) -> str: ...
    def special_env(
        self,
        name: str,
        has_kdc_conf: bool,
        krb5_conf: Dict[str, str] | None = ...,
        kdc_conf: Dict[str, str] | None = ...,
    ) -> Dict[str, str]: ...
    def kill_daemons(self) -> None: ...



# Copyright: (c) 2022 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from libc.stdlib cimport free

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    # Heimdal and MIT differ in their implementations
    """
    krb5_error_code krb5_enctype_to_string_generic(
        krb5_context context,
        krb5_enctype enctype,
        char **buffer
    )
    {
    // Heimdal takes in a context and sets an output pointer
    #if defined(HEIMDAL_XFREE)
        return krb5_enctype_to_string(context, enctype, buffer);
    #else
        char *tmp = malloc(100);
        if (tmp == NULL)
        {
            return ENOMEM;
        }
        *buffer = tmp;

        return krb5_enctype_to_string(enctype, tmp, 100);
    #endif
    }

    krb5_error_code krb5_string_to_enctype_generic(
        krb5_context context,
        char *string,
        krb5_enctype *enctypep
    )
    {
    // MIT does not have a context overload
    #if defined(HEIMDAL_XFREE)
        return krb5_string_to_enctype(context, string, enctypep);
    #else
        return krb5_string_to_enctype(string, enctypep);
    #endif
    }
    """

    krb5_error_code krb5_enctype_to_string_generic(
        krb5_context context,
        krb5_enctype enctype,
        char **buffer,
    ) nogil

    krb5_error_code krb5_string_to_enctype_generic(
        krb5_context context,
        char *string,
        krb5_enctype *enctypep,
    ) nogil


def enctype_to_string(
    Context context not None,
    krb5_enctype enctype,
) -> str:
    cdef krb5_error_code err = 0
    cdef char *buffer = NULL

    err = krb5_enctype_to_string_generic(context.raw, enctype, &buffer)
    if err:
        raise Krb5Error(context, err)

    try:
        return buffer.decode("utf-8")
    finally:
        free(buffer)


def string_to_enctype(
    Context context not None,
    str string,
) -> int:
    cdef krb5_enctype enctype = 0
    cdef krb5_error_code err = 0
    b_string = string.encode("utf-8")
    cdef char *string_ptr = b_string

    err = krb5_string_to_enctype_generic(context.raw, string_ptr, &enctype)
    if err:
        raise Krb5Error(context, err)

    return enctype



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os
import time

import k5test
import pytest

import krb5


def test_init_context() -> None:
    context = krb5.init_context()
    assert context is not None
    assert isinstance(context, krb5.Context)
    assert str(context) == "Krb5Context"


def test_set_default_realm(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()

    default = krb5.get_default_realm(ctx)
    assert default == realm.realm.encode()

    krb5.set_default_realm(ctx, b"NEW.REALM")
    default = krb5.get_default_realm(ctx)
    assert default == b"NEW.REALM"

    krb5.set_default_realm(ctx, None)
    default = krb5.get_default_realm(ctx)
    assert default == realm.realm.encode()


def test_set_real_time() -> None:
    ctx = krb5.init_context()

    diff = krb5.timeofday(ctx) - time.time()
    assert diff > -5
    assert diff < 5

    now = krb5.us_timeofday(ctx)
    diff = now[0] * 1000000 + now[1] - int(time.time() * 1e6)
    assert diff > -5000000
    assert diff < 5000000

    krb5.set_real_time(ctx, int(time.time()) + 100, 0)
    diff = krb5.timeofday(ctx) - time.time()
    assert diff > 95
    assert diff < 105

    krb5.set_real_time(ctx, int(time.time()), -1)
    diff = krb5.timeofday(ctx) - time.time()
    assert diff > -5
    assert diff < 5


@pytest.mark.requires_api("get_time_offsets")
def test_get_time_offsets() -> None:
    ctx = krb5.init_context()

    sec, usec = krb5.get_time_offsets(ctx)
    assert sec == 0
    assert usec == 0

    krb5.set_real_time(ctx, int(time.time()) + 100, 0)
    sec, usec = krb5.get_time_offsets(ctx)
    assert sec > 95
    assert sec < 105

    krb5.set_real_time(ctx, int(time.time()), -1)
    sec, usec = krb5.get_time_offsets(ctx)
    assert sec > -5
    assert sec < 5


@pytest.mark.requires_api("init_secure_context")
def test_init_secure_context() -> None:
    context = krb5.init_secure_context()
    assert context is not None
    assert isinstance(context, krb5.Context)
    assert str(context) == "Krb5Context"



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._ccache import CCache
from krb5._context import Context
from krb5._creds import Creds
from krb5._creds_opt import GetInitCredsOpt
from krb5._principal import Principal

class EtypeInfo(typing.NamedTuple):
    etype: int
    salt: typing.Optional[bytes]
    s2kparams: typing.Optional[bytes]

def get_validated_creds(
    context: Context,
    client: Principal,
    ccache: CCache,
    in_tkt_service: typing.Optional[bytes] = None,
) -> Creds:
    """Get validated credentials from the KDC for a postdated ticket.

    Args:
        context: Krb5 context.
        client: Client principal name.
        ccache: The cache to get the existing credentials from.
        in_tkt_service: Server principal string or None.
    """

def get_etype_info(
    context: Context,
    principal: Principal,
    opt: typing.Optional[GetInitCredsOpt] = None,
) -> EtypeInfo:
    """Retrieve the enctype, salt and s2kparams for a principal from the KDC.

    Args:
        context: Krb5 context.
        principal: Principal to fetch the information for.
        opt: Options to use (e.g. for FAST armoring).

    Returns:
        A named tuple containing the enctype, the salt and the s2kparams.

        If the KDC provides no etype-info, the returned salt is None.

        If there are no s2kparams in the provided etype-info, s2kparams is None.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum
import typing

from krb5._ccache import CCache
from krb5._context import Context
from krb5._creds_opt import GetInitCredsOpt
from krb5._keyblock import KeyBlock
from krb5._kt import KeyTab
from krb5._principal import Principal

class TicketTimes(typing.NamedTuple):
    authtime: int
    starttime: int
    endtime: int
    renew_till: int

class TicketFlags(enum.IntFlag):
    """Kerberos ticket flags.

    See <https://www.rfc-editor.org/rfc/rfc4120.html#page-68> for the definition
    of the flags.

    The enc_pa_rep flag is defined by
    <https://www.rfc-editor.org/rfc/rfc6806.html#section-11>.

    The anonymous flag is defined by
    <https://www.rfc-editor.org/rfc/rfc6112.html#section-3>.
    """

    reserved = 1 << 0
    forwardable = 1 << 1
    forwarded = 1 << 2
    proxiable = 1 << 3
    proxy = 1 << 4
    may_postdate = 1 << 5
    postdated = 1 << 6
    invalid = 1 << 7
    renewable = 1 << 8
    initial = 1 << 9
    pre_authent = 1 << 10
    hw_authent = 1 << 11
    transited_policy_checked = 1 << 12
    ok_as_delegate = 1 << 13
    enc_pa_rep = 1 << 15
    anonymous = 1 << 16

class Creds:
    """Kerberos Credentials object.

    This class represents Kerberos credentials.

    Args:
        context: Krb5 context.
    """

    @property
    def client(self) -> Principal:
        """Client's principal identifier."""
    @property
    def server(self) -> Principal:
        """Server's principal identifier."""
    @property
    def keyblock(self) -> KeyBlock:
        """Session encryption key info."""
    @property
    def times(self) -> TicketTimes:
        """Lifetime info."""
    @property
    def ticket_flags_raw(self) -> int:
        """Flags in ticket, as returned by libkrb5."""
    @property
    def ticket_flags(self) -> TicketFlags:
        """Flags in ticket, converted to a representation where the first flag is in the lowermost bit."""
    # @property
    # def addresses(self) -> Address:
    #     """Addrs in ticket."""
    @property
    def ticket(self) -> bytes:
        """Ticket string itself."""
    @property
    def second_ticket(self) -> bytes:
        """second ticket, if related to ticket (via DUPLICATE-SKEY or ENC-TKT-IN-SKEY)"""
    # @property
    # def authdata(self) -> Authdata:
    #     """authorization data"""

class InitCredsContext:
    """Kerberos Initial Credentials object.

    This class represents Kerberos context for acquiring initial credentials.

    Args:
        context: Krb5 context.
    """

class Krb5Prompt:
    """Base class used to prompt the user for input.

    This is the base class the caller should use to create a prompter class
    that is passed into :meth:`get_init_creds_password` and
    :meth:`init_creds_init`. It is used to handle prompts from the Kerberos
    API, such as changing a password or password input.

    At a minimum the `prompt` function should be implemented to return the
    bytes for the prompt.
    """

    def init(
        self,
        name: typing.Optional[bytes],
        banner: typing.Optional[bytes],
        num_prompts: int,
    ) -> None:
        """Initialise the prompt.

        Initialises the prompt by passing in the optional name, banner, and the
        number of prompts expected.

        Args:
            name: Optional name for the prompt session which should be
                displayed to the user.
            banner: Optional message that indicates what sort of input is
                expected from the user which should be displayed to the user.
            num_prompts: Number of times :meth:`prompt` is to be called for
                this prompt session.
        """
    def prompt(
        self,
        msg: bytes,
        hidden: bool,
    ) -> bytes:
        """Prompt user for a response.

        Prompt the user for a response.

        Args:
            msg: The message to display for this prompt.
            hidden: Set to ``True`` to indicate the response should not be
                printed back to the user.

        Returns:
            bytes: The response for the prompt.
        """

def get_init_creds_keytab(
    context: Context,
    client: Principal,
    k5_gic_options: GetInitCredsOpt,
    keytab: KeyTab,
    start_time: int = 0,
    in_tkt_service: typing.Optional[bytes] = None,
) -> Creds:
    """Get initial credentials using a key table.

    Requests the KDC for credential using a client key stored in the key table
    specified.

    Args:
        context: Krb5 context.
        client: The client principal the credentials are for.
        k5_gic_options: The initial credentials options.
        keytab: The keytab to use when getting the credential.
        start_time: Time when the ticket becomes valid, 0 for now.
        in_tkt_service: The service name of the initial credentials.

    Returns:
        Creds: The retrieved credentials.
    """

def get_init_creds_password(
    context: Context,
    client: Principal,
    k5_gic_options: GetInitCredsOpt,
    password: typing.Optional[bytes] = None,
    start_time: int = 0,
    in_tkt_service: typing.Optional[bytes] = None,
    prompter: typing.Optional[Krb5Prompt] = None,
) -> Creds:
    """Get initial credential using a password.

    Requests the KDC for credentials using a password.

    Args:
        context: Krb5 context.
        client: The client principal the credentials are for.
        k5_gic_options: The initial credentials options.
        password: The password to use - set to ``None`` to use the prompter.
        start_time: Time when the ticket becomes valid, 0 for now.
        in_tkt_service: The service name of the initial credentials.
        prompter: The prompter instance that will handle interactive prompts.

    Returns:
        Creds: The retrieved credentials.
    """

def init_creds_get(
    context: Context,
    ctx: InitCredsContext,
) -> None:
    """Acquire credentials using initial creds context.

    Obtains credentials using the initial creds context from
    :meth:`init_creds_init`. The credentials can be retrieved with
    :meth:`init_creds_get_creds`.

    Args:
        context: Krb5 context.
        ctx: Initial credentials context.
    """

def init_creds_get_creds(
    context: Context,
    ctx: InitCredsContext,
) -> Creds:
    """Retreived acquired creds from an initial creds context.

    Gets the acquired creds from a successful call of :meth:`init_creds_get`.

    Args:
        context: Krb5 context.
        ctx: Initial credentials context.

    Returns:
        Creds: The acquired credentials.
    """

def init_creds_init(
    context: Context,
    client: Principal,
    k5_gic_options: typing.Optional[GetInitCredsOpt] = None,
    start_time: int = 0,
    prompter: typing.Optional[Krb5Prompt] = None,
) -> InitCredsContext:
    """Get initial acquiring credential context.

    Creates a new context used for acquiring initial credentials.

    Args:
        context: Krb5 context.
        client: The client principal the credentials are for.
        k5_gic_options: The initial credentials options.
        start_time: Time when the ticket becomes valid, 0 for now.
        prompter: The prompter instance that will handle interactive prompts.

    Returns:
        InitCredsContext: The retrieved acquiring initial credentials context.
    """

def init_creds_set_keytab(
    context: Context,
    ctx: InitCredsContext,
    keytab: KeyTab,
) -> None:
    """Set the keytab used for acquiring initial credentials.

    Sets the keytab used to contruct the client key for an initial
    credentials request. See :meth:`get_init_creds_keytab` for a simpler
    way to get credentials with a keytab.

    Args:
        context: Krb5 context.
        ctx: Initial credentials context.
        keytab: The keytab to set.
    """

def init_creds_set_password(
    context: Context,
    ctx: InitCredsContext,
    password: bytes,
) -> None:
    """Set the password used for acquiring initial credentials.

    Sets the password used to contruct the client key for an initial
    credentials request. See :meth:`get_init_creds_password` for a simpler
    way to get credentials with a password.

    Args:
        context: Krb5 context.
        ctx: Initial credentials context.
        password: The password to set.
    """

def get_renewed_creds(
    context: Context,
    client: Principal,
    ccache: CCache,
    in_tkt_service: typing.Optional[bytes] = None,
) -> Creds:
    """Get renewed credentials from the KDC.

    Args:
        context: Krb5 context.
        client: Client principal name.
        ccache: The cache to get the existing credentials from.
        in_tkt_service: Server principal string or None.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum
import typing

from krb5._context import Context
from krb5._creds import Creds
from krb5._principal import Principal

class CredentialsRetrieveFlags(enum.IntEnum):
    """Flags used to control :meth:`cc_retrieve_cred` and  :meth:`cc_remove_cred`."""

    none: CredentialsRetrieveFlags = ...  #: No matching flags set
    match_times: CredentialsRetrieveFlags = ...  #: The requested lifetime must be at least as great as specified
    match_is_skey: CredentialsRetrieveFlags = ...  #: The is_skey field must match exactly
    match_flags: CredentialsRetrieveFlags = ...  #: All the flags set in the match credentials must be set
    match_times_exact: CredentialsRetrieveFlags = ...  #: All the time fields must match exactly
    match_flags_exact: CredentialsRetrieveFlags = ...  #: All the flags must match exactly
    match_authdata: CredentialsRetrieveFlags = ...  #: The authorization data must match
    match_srv_nameonly: CredentialsRetrieveFlags = ...  #: Only the name portion of the principal name must match
    match_2nd_tkt: CredentialsRetrieveFlags = ...  #: The second ticket must match
    match_keytype: CredentialsRetrieveFlags = ...  #: The encryption key type must match
    supported_ktypes: CredentialsRetrieveFlags = ...  #: The supported key types must match

class CCache:
    """Kerberos CCache

    This class represents a Credential Cache object.

    Args:
        context: Krb5 context.
    """

    def __iter__(self) -> typing.Iterator[Creds]:
        """Iterate credentials in a ccache."""
    @property
    def addr(self) -> typing.Optional[int]:
        """The raw krb5_ccache pointer address of this credential cache."""
    @property
    def name(self) -> typing.Optional[bytes]:
        """The name/residual of the credential cache."""
    @property
    def principal(self) -> typing.Optional[Principal]:
        """Default client principal of the credential cache."""
    @property
    def cache_type(self) -> typing.Optional[bytes]:
        """The type of the credential cache."""

def cc_default(
    context: Context,
) -> CCache:
    """Resolve the default credential cache name.

    Create a handle to the default credential cache as given by
    :meth:`cc_default_name()`.

    Args:
        context: Krb5 context.

    Returns:
        CCache: The opened credential cache.
    """

def cc_default_name(
    context: Context,
) -> bytes:
    """Get the name of the default credential cache.

    The name of the default credential cache. This value consults a wide range
    of sources such as the ``KRB5CCNAME`` env var, config settings, build time
    variables, etc.

    Args:
        context: Krb5 context.

    Returns:
        bytes: The name of the default credential cache.
    """

def cc_destroy(
    context: Context,
    cache: CCache,
) -> None:
    """Destroy a credential cache.

    Destroys the contents of the cache and closes the handle to it.

    Args:
        context: Krb5 context.
        cache: The credential cache to destroy.
    """

def cc_get_name(
    context: Context,
    cache: CCache,
) -> bytes:
    """Retrieve the name of a credential cache.

    Gets the name of the credential, but not the type of a credential cache.

    Args:
        context: Krb5 context.
        cache: The credential cache to query.

    Returns:
        bytes: The credential cache name.
    """

def cc_get_principal(
    context: Context,
    cache: CCache,
) -> Principal:
    """Get the default principal of a credential cache.

    Gets the default principal of a credential as set by
    :meth:`cc_initialize()`.

    Args:
        context: Krb5 context.
        cache: The credential cache to query.

    Returns:
        Principal: The default principal of the cache.
    """

def cc_get_type(
    context: Context,
    cache: CCache,
) -> bytes:
    """Retrieve the type of a credential cache.

    Gets the credential cache type.

    Args:
        context: Krb5 context.
        cache: The credential cache to query.

    Returns:
        bytes: The credential cache type.
    """

def cc_initialize(
    context: Context,
    cache: CCache,
    principal: Principal,
) -> None:
    """Initialize a credential cache.

    Destroy any existing contents of a cache and initialize it for the default
    principal specified.

    Args:
        context: Krb5 context.
        cache: The cache to destroy and recreate.
        principal: The default principal of the cache.
    """

def cc_new_unique(
    context: Context,
    cred_type: bytes,
    hint: typing.Optional[bytes] = None,
) -> CCache:
    """Create a new unique credential cache.

    Create a new credential cache of the specified type with a unique name.

    Args:
        context: Krb5 context.
        cred_type: The credential cache type to create.
        hint: Unused.

    Returns:
        CCache: The created credential cache.
    """

def cc_remove_cred(
    context: Context,
    cache: CCache,
    flags: typing.Union[int, CredentialsRetrieveFlags],
    creds: Creds,
) -> None:
    """Remove matching credentials from a credential cache.

    Remove all credential which match creds according to flags from the
    credential cache.

    Args:
        context: Krb5 context.
        cache: The credential cache to store the creds into.
        flags: The flags describing how to perform the matching.
        creds: The credentials to match against.
    """

def cc_resolve(
    context: Context,
    name: bytes,
) -> CCache:
    """Resolve a credential cache name.

    Resolves/opens a credential cache by the name specified. The name should be
    in the form ``type:residual`` where the type is known to the krb5 library
    being called. If the name does not contain a colon then it is interpreted
    as a file name (``FILE:*``).

    Args:
        context: Krb5 context.
        name: The name of the cache to resolve.

    Returns:
        CCache: The credential cache that was resolved.
    """

def cc_retrieve_cred(
    context: Context,
    cache: CCache,
    flags: typing.Union[int, CredentialsRetrieveFlags],
    mcreds: Creds,
) -> Creds:
    """Retrieve matching credentials from a credential cache.

    Retrieve all credential which match creds according to flags from the
    credential cache.

    Args:
        context: Krb5 context.
        cache: The credential cache to store the creds into.
        flags: The flags describing how to perform the matching.
        mcreds: The credentials to match against.

    Returns:
        Creds: The matching credentials.
    """

def cc_set_default_name(
    context: Context,
    name: typing.Optional[bytes],
) -> None:
    """Set the default ccache name.

    Set the default credential cache name to the name specified for future
    operations using the context. If name is `None` or an empty byte string
    this will clear any previous application-set default name and forget any
    cached value of the default name for the context.

    Args:
        context: Krb5 context.
        name: The default credential name or `None` to reset back to the config
            defaults.
    """

def cc_store_cred(
    context: Context,
    cache: CCache,
    creds: Creds,
) -> None:
    """Store credentials in a credential cache.

    Stores the credentials into the credential cache specified.

    Args:
        context: Krb5 context.
        cache: The credential cache to store the creds into.
        creds: The credentials to store.
    """

def cc_switch(
    context: Context,
    cache: CCache,
) -> None:
    """Switch primary cache in a collection.

    If the type of cache supports it, set the cache to be the primary
    credential cache for the collection it belongs to.

    Args:
        context: Krb5 context.
        cache: The credential cache to set as the primary in its collection.
    """

def cc_get_config(
    context: Context,
    cache: CCache,
    principal: typing.Optional[Principal],
    key: bytes,
) -> bytes:
    """Get a configuration value from a credential cache.

    Args:
        context: Krb5 context.
        cache: The credential cache to get the data from.
        principal: The principal to get the configuration for or None for global values.
        key: Name of the variable.
    """

def cc_set_config(
    context: Context,
    cache: CCache,
    principal: typing.Optional[Principal],
    key: bytes,
    data: bytes,
) -> bytes:
    """Store a configuration value in a credential cache.

    Args:
        context: Krb5 context.
        cache: The credential cache to store the data in.
        principal: The principal to set the configuration for or None for global values.
        key: Name of the variable.
        data: Data to store or None to remove.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._context cimport Context
from krb5._exceptions import Krb5Error
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    krb5_error_code krb5_init_secure_context(
        krb5_context *context,
    ) nogil

    krb5_error_code krb5_get_time_offsets(
        krb5_context context,
        krb5_timestamp *seconds,
        int32_t *microseconds
    ) nogil


def init_secure_context() -> Context:
    context = Context()
    krb5_init_secure_context(&context.raw)

    return context


def get_time_offsets(
    Context context not None,
) -> typing.Tuple[int, int]:
    cdef krb5_error_code = 0

    cdef krb5_timestamp seconds
    cdef int32_t microseconds
    err = krb5_get_time_offsets(context.raw, &seconds, &microseconds)
    if err:
        raise Krb5Error(context, err)

    return seconds, microseconds



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._ccache import CCache
from krb5._context import Context

def cc_dup(
    context: Context,
    cache: CCache,
) -> CCache:
    """Duplicate ccache handle.

    Create a new handle referring to the same cache referenced. The new cache
    can be closed independently.

    Args:
        context: Krb5 context.
        cache: The credential cache to duplicate.

    Returns:
        CCache: The duplicated ccache.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._krb5_types cimport *
from krb5._kt cimport KeyTab


cdef extern from "python_krb5.h":
    krb5_error_code krb5_kt_client_default(
        krb5_context context,
        krb5_keytab *keytab_out,
    ) nogil

    krb5_error_code krb5_kt_dup(
        krb5_context context,
        krb5_keytab in_kt,
        krb5_keytab *out,
    ) nogil


def kt_client_default(
    Context context not None,
) -> KeyTab:
    kt = KeyTab(context)
    cdef krb5_error_code err = 0

    err = krb5_kt_client_default(context.raw, &kt.raw)
    if err:
        raise Krb5Error(context, err)

    return kt


def kt_dup(
    Context context not None,
    KeyTab keytab not None,
) -> KeyTab:
    out_kt = KeyTab(context)
    cdef krb5_error_code err = 0

    err = krb5_kt_dup(context.raw, keytab.raw, &out_kt.raw)
    if err:
        raise Krb5Error(context, err)

    return out_kt



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import collections
import typing

from krb5._exceptions import Krb5Error

from krb5._ccache cimport CCache
from krb5._context cimport Context
from krb5._creds cimport Creds
from krb5._creds_opt cimport GetInitCredsOpt
from krb5._krb5_types cimport *
from krb5._principal cimport Principal


cdef extern from "python_krb5.h":
    krb5_error_code krb5_get_validated_creds(
        krb5_context context,
        krb5_creds *creds,
        krb5_principal client,
        krb5_ccache ccache,
        const char *in_tkt_service,
    ) nogil

    krb5_error_code krb5_get_etype_info(
        krb5_context context,
        krb5_principal principal,
        krb5_get_init_creds_opt *opt,
        krb5_enctype *enctype_out,
        krb5_data *salt_out,
        krb5_data *s2kparams_ou,
    ) nogil

EtypeInfo = collections.namedtuple('EtypeInfo', [
    'etype',
    'salt',
    's2kparams',
])

def get_validated_creds(
    Context context not None,
    Principal client not None,
    CCache ccache not None,
    const unsigned char[:] in_tkt_service = None,
) -> Creds:
    creds = Creds(context)
    cdef krb5_error_code err = 0

    cdef const char *in_tkt_service_ptr = NULL
    if in_tkt_service is not None and len(in_tkt_service):
        in_tkt_service_ptr = <const char*>&in_tkt_service[0]

    err = krb5_get_validated_creds(
        context.raw,
        creds.get_pointer(),
        client.raw,
        ccache.raw,
        in_tkt_service_ptr)
    if err:
        raise Krb5Error(context, err)

    creds.free_contents = 1

    return creds

def get_etype_info(
    Context context not None,
    Principal principal not None,
    GetInitCredsOpt opt = None,
) -> EtypeInfo:
    cdef krb5_error_code err = 0

    cdef krb5_get_init_creds_opt *options = NULL
    if opt:
        options = opt.raw

    cdef krb5_enctype enctype
    cdef krb5_data salt
    cdef krb5_data s2kparams

    with nogil:
        err = krb5_get_etype_info(
            context.raw,
            principal.raw,
            options,
            &enctype,
            &salt,
            &s2kparams,
        )
    if err:
        raise Krb5Error(context, err)

    cdef size_t length
    cdef char *value

    pykrb5_get_krb5_data(&salt, &length, &value)
    if length == 0:
        if enctype == 0:  # ENCTYPE_NULL
            # If enctype is ENCTYPE_NULL and the salt is empty this means that
            # the KDC provided no etype-info. Return None in this case.
            salt_bytes = None
        else:
            salt_bytes = b""
    else:
        salt_bytes = <bytes>value[:length]
    pykrb5_free_data_contents(context.raw, &salt)

    pykrb5_get_krb5_data(&s2kparams, &length, &value)
    if length == 0:
        # This means that the KDC provided an etype-info without s2kparams
        # (PA-ETYPE-INFO instead of PA-ETYPE-INFO2).
        # Return None instead of b'' in this case so that the value can be
        # passed directly to krb.c_string_to_key()
        s2kparams_bytes = None
    else:
        s2kparams_bytes = <bytes>value[:length]
    pykrb5_free_data_contents(context.raw, &s2kparams)

    return EtypeInfo(enctype, salt_bytes, s2kparams_bytes)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._context cimport Context
from krb5._krb5_types cimport *
from krb5._principal cimport Principal


cdef extern from "python_krb5.h":
    const char *krb5_principal_get_realm(
        krb5_context context,
		krb5_const_principal principal
    ) nogil


def principal_get_realm(
    Context context not None,
    Principal principal not None,
) -> bytes:
    return <bytes>krb5_principal_get_realm(context.raw, principal.raw)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum
import typing

from cpython cimport array
from libc.stdint cimport int32_t, uintptr_t

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    # krb5_free_unparsed_name is deprecated in Heimdal - use krb5_xfree instead
    """
    #if defined(HEIMDAL_XFREE)
    #include "krb5_asn1.h"
    #endif

    void krb5_free_unparsed_name_generic(krb5_context context, char *val)
    {
    #if defined(HEIMDAL_XFREE)
        krb5_xfree(val);
    #else
        krb5_free_unparsed_name(context, val);
    #endif
    }

    void pykrb5_principal_get(
        krb5_principal principal,
        size_t *realm_name_length,
        char **realm_name,
        size_t *component_count,
        int32_t *type
    )
    {
    #if defined(HEIMDAL_XFREE)
        if (realm_name_length != NULL) *realm_name_length = strlen(principal->realm);
        if (realm_name != NULL) *realm_name = principal->realm;
        if (component_count != NULL) *component_count = principal->name.name_string.len;
        if (type != NULL) *type = principal->name.name_type;
    #else
        if (realm_name_length != NULL) *realm_name_length = principal->realm.length;
        if (realm_name != NULL) *realm_name = principal->realm.data;
        if (component_count != NULL) *component_count = principal->length;
        if (type != NULL) *type = principal->type;
    #endif
    }

    int pykrb5_principal_set(
        krb5_principal principal,
        size_t realm_name_length,
        const char *realm_name,
        size_t component_count
    )
    {
        memset(principal, 0, sizeof(*principal));

    #if defined(HEIMDAL_XFREE)
        principal->realm = (char *)realm_name;
        principal->name.name_type = 0;
        principal->name.name_string.len = component_count;

        principal->name.name_string.val = malloc(component_count * sizeof(krb5_data));
        if (!principal->name.name_string.val)
            return 0;
    #else
        principal->realm.length = realm_name_length;
        principal->realm.data = (char *)realm_name;
        principal->length = component_count;
        principal->type = 0;

        principal->data = malloc(component_count * sizeof(krb5_data));
        if (!principal->data)
            return 0;
    #endif

        return 1;
    }

    void pykrb5_principal_set_type(
        krb5_principal principal,
        int32_t type
    )
    {
    #if defined(HEIMDAL_XFREE)
        principal->name.name_type = type;
    #else
        principal->type = type;
    #endif
    }

    void pykrb5_principal_get_component(
        krb5_principal principal,
        size_t pos,
        size_t *component_length,
        char **component
    )
    {
    #if defined(HEIMDAL_XFREE)
        if (component_length != NULL) *component_length = strlen(principal->name.name_string.val[pos]);
        if (component != NULL) *component = principal->name.name_string.val[pos];
    #else
        if (component_length != NULL) *component_length = principal->data[pos].length;
        if (component != NULL) *component = principal->data[pos].data;
    #endif
    }

    void pykrb5_principal_set_component(
        krb5_principal principal,
        size_t pos,
        size_t component_length,
        const char *component
    )
    {
    #if defined(HEIMDAL_XFREE)
        principal->name.name_string.val[pos] = (char *)component;
    #else
        principal->data[pos].length = component_length;
        principal->data[pos].data = (char *)component;
    #endif
    }

    void pykrb5_principal_set_free(
        krb5_principal principal
    )
    {
    #if defined(HEIMDAL_XFREE)
        free(principal->name.name_string.val);
    #else
        free(principal->data);
    #endif
    }
    """

    krb5_error_code krb5_copy_principal(
        krb5_context context,
        krb5_const_principal inprinc,
        krb5_principal *outprinc,
    ) nogil

    void krb5_free_principal(
        krb5_context context,
        krb5_principal val,
    ) nogil

    # See inline C code
    void krb5_free_unparsed_name_generic(
        krb5_context context,
        char *val,
    ) nogil

    krb5_error_code krb5_parse_name_flags(
        krb5_context context,
        const char *name,
        int flags,
        krb5_principal *principal_out,
    ) nogil

    krb5_error_code krb5_unparse_name_flags(
        krb5_context context,
        krb5_const_principal principal,
        int flags,
        char **name,
    ) nogil

    int32_t KRB5_PRINCIPAL_PARSE_NO_REALM
    int32_t KRB5_PRINCIPAL_PARSE_REQUIRE_REALM
    int32_t KRB5_PRINCIPAL_PARSE_ENTERPRISE
    int32_t KRB5_PRINCIPAL_PARSE_IGNORE_REALM

    int32_t KRB5_PRINCIPAL_UNPARSE_SHORT
    int32_t KRB5_PRINCIPAL_UNPARSE_NO_REALM
    int32_t KRB5_PRINCIPAL_UNPARSE_DISPLAY

    void pykrb5_principal_get(
        krb5_principal principal,
        size_t *realm_name_length,
        const char **realm_name,
        size_t *component_count,
        int32_t *type
    ) nogil

    int pykrb5_principal_set(
        krb5_principal principal,
        size_t realm_name_length,
        const char *realm_name,
        size_t component_count,
    ) nogil

    void pykrb5_principal_set_type(
        krb5_principal principal,
        int32_t type
    ) nogil

    void pykrb5_principal_get_component(
        krb5_principal principal,
        size_t pos,
        size_t *component_length,
        const char **component
    ) nogil

    void pykrb5_principal_set_component(
        krb5_principal principal,
        size_t pos,
        size_t component_length,
        const char *component,
    )

    void pykrb5_principal_set_free(
        krb5_principal principal,
    )


class PrincipalParseFlags(enum.IntEnum):
    none = 0
    no_realm = KRB5_PRINCIPAL_PARSE_NO_REALM
    require_realm = KRB5_PRINCIPAL_PARSE_REQUIRE_REALM
    enterprise = KRB5_PRINCIPAL_PARSE_ENTERPRISE
    ignore_realm = KRB5_PRINCIPAL_PARSE_IGNORE_REALM


class PrincipalUnparseFlags(enum.IntEnum):
    none = 0
    short = KRB5_PRINCIPAL_UNPARSE_SHORT
    no_realm = KRB5_PRINCIPAL_UNPARSE_NO_REALM
    display = KRB5_PRINCIPAL_UNPARSE_DISPLAY


class NameType(enum.IntEnum):
    # See also https://github.com/krb5/krb5-assignments/blob/master/name-type

    # https://www.rfc-editor.org/rfc/rfc4120.html#section-7.5.8
    unknown = 0
    principal = 1
    srv_inst = 2
    srv_hst = 3
    srv_xhst = 4
    uid = 5
    x500_principal = 6
    smtp_name = 7
    enterprise_principal = 10
    # https://www.rfc-editor.org/rfc/rfc6111.html#section-3.1
    wellknown = 11
    # Not standardized, supported by MIT and Heimdal
    ms_principal = -128
    ms_principal_and_id = -129
    ent_principal_and_id = -130

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        if not isinstance(value, int):
            return None
        value = int(value)

        new_member = int.__new__(cls, value)
        new_member._name_ = f"Unknown_NameType_{str(value).replace('-', 'm')}"
        new_member._value_ = value
        return cls._value2member_map_.setdefault(value, new_member)


cdef class Principal:
    """Kerberos Principal object.

    This class represents a Kerberos principal.

    Args:
        context: Krb5 context.
    """
    # cdef Context ctx
    # cdef krb5_principal raw
    # cdef int needs_free
    # cdef int _parse_flags

    def __cinit__(Principal self, Context context, flags, int needs_free=1):
        self.ctx = context
        self.raw = NULL
        self.needs_free = needs_free
        self._parse_flags = flags

    def __copy__(Principal self):
        return copy_principal(self.ctx, self)

    def __dealloc__(Principal self):
        if self.raw and self.needs_free:
            krb5_free_principal(self.ctx.raw, self.raw)
            self.raw = NULL

    @property
    def addr(Principal self) -> typing.Optional[int]:
        if self.raw:
            return <uintptr_t>self.raw

    @property
    def name(Principal self) -> typing.Optional[bytes]:
        if self.raw:
            # Heimdal fails to unparse a no_realm principal if the no_realm unparse flags aren't used.
            flags = PrincipalUnparseFlags.none
            if (
                self._parse_flags & PrincipalParseFlags.no_realm or
                self._parse_flags & PrincipalParseFlags.ignore_realm
            ):
                flags = PrincipalUnparseFlags.no_realm

            return unparse_name_flags(self.ctx, self, flags=flags)

    def __repr__(Principal self) -> str:
        name = self.name
        return f"Principal({name.decode('utf-8') if name else 'NULL'})"

    def __str__(Principal self) -> str:
        name = self.name
        return name.decode('utf-8') if name else 'NULL'

    @property
    def realm(Principal self) -> bytes:
        cdef size_t length
        cdef char *value

        if not self.raw:
            raise ValueError("Attempting to access property of NULL principal")

        pykrb5_principal_get(self.raw, &length, &value, NULL, NULL)

        if length == 0:
            return b""
        else:
            return value[:length]

    @property
    def components(Principal self) -> typing.List[bytes]:
        cdef size_t component_count

        cdef size_t length
        cdef char *value

        if not self.raw:
            raise ValueError("Attempting to access property of NULL principal")

        pykrb5_principal_get(self.raw, NULL, NULL, &component_count, NULL)

        components = []
        for pos in range(component_count):
            pykrb5_principal_get_component(self.raw, pos, &length, &value)

            if length == 0:
                component = b""
            else:
                component = value[:length]

            components.append(component)

        return components

    @property
    def type(Principal self) -> NameType:
        cdef int32_t type

        if not self.raw:
            raise ValueError("Attempting to access property of NULL principal")

        pykrb5_principal_get(self.raw, NULL, NULL, NULL, &type)
        return NameType(type)

    @type.setter
    def type(Principal self, int32_t value):
        if not self.raw:
            raise ValueError("Attempting to access property of NULL principal")

        pykrb5_principal_set_type(self.raw, value)


def copy_principal(
    Context context not None,
    Principal principal not None,
) -> Principal:
    out = Principal(context, principal._parse_flags)
    cdef krb5_error_code err = 0

    err = krb5_copy_principal(context.raw, principal.raw, &out.raw)
    if err:
        raise Krb5Error(context, err)

    return out


def parse_name_flags(
    Context context not None,
    const unsigned char[:] name not None,
    int flags=PrincipalParseFlags.none,
) -> Principal:
    principal = Principal(context, flags)
    cdef krb5_error_code err = 0

    cdef const char *name_ptr = NULL
    if name is not None and len(name):
        name_ptr = <const char*>&name[0]
    else:
        raise ValueError("Principal must be set")

    err = krb5_parse_name_flags(context.raw, name_ptr, flags, &principal.raw)
    if err:
        raise Krb5Error(context, err)

    return principal


def unparse_name_flags(
    Context context not None,
    Principal principal not None,
    int flags=PrincipalUnparseFlags.none,
) -> bytes:
    cdef krb5_error_code err = 0
    cdef char *name = NULL

    err = krb5_unparse_name_flags(context.raw, principal.raw, flags, &name)
    if err:
        raise Krb5Error(context, err)

    try:
        return <bytes>name
    finally:
        krb5_free_unparsed_name_generic(context.raw, name)


# Note: build_principal() does not actually call krb5_build_principal() because
# this would require passing vararg parameters and because
# krb5_build_principal() cannot handle NUL bytes in the strings.
# Instead, the principal is built manually, and then duplicated with
# copy_principal().
def build_principal(
    Context context not None,
    const unsigned char[:] realm not None,
    components: typing.Iterable[bytes],
) -> Principal:
    component_list = list(components)
    component_count = len(component_list)

    # Guess the component count, similar to the MIT implementation starting with 1.12
    if component_count == 2 and component_list[0] == b'krbtgt':
        inferred_principal_type = NameType.srv_inst
    elif component_count >= 2 and component_list[0] == b'WELLKNOWN':
        inferred_principal_type = NameType.wellknown
    else:
        inferred_principal_type = NameType.principal

    cdef krb5_error_code err = 0

    cdef krb5_principal_data raw
    if pykrb5_principal_set(&raw, len(realm), <const char*>&realm[0], component_count) == 0:
        raise MemoryError()
    try:
        for pos in range(component_count):
            component = component_list[pos]
            pykrb5_principal_set_component(&raw, pos, len(component), <char*>component)

        out = Principal(context, 0)

        err = krb5_copy_principal(context.raw, &raw, &out.raw)
        if err:
            raise Krb5Error(context, err)

        out.type = inferred_principal_type

        return out
    finally:
        pykrb5_principal_set_free(&raw)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._exceptions import Krb5Error

from krb5._context cimport Context
from krb5._creds_opt cimport GetInitCredsOpt
from krb5._krb5_types cimport *


cdef extern from "python_krb5.h":
    void krb5_get_init_creds_opt_set_default_flags(
        krb5_context context,
        const char *appname,
        const char *realm,
        krb5_get_init_creds_opt *opt,
    ) nogil


def get_init_creds_opt_set_default_flags(
    Context context not None,
    GetInitCredsOpt opt not None,
    const unsigned char[:] appname = None,
    const unsigned char[:] realm = None,
) -> None:
    cdef const char *appname_ptr = NULL
    if appname is not None and len(appname):
        appname_ptr = <const char*>&appname[0]

    cdef const char *realm_ptr = NULL
    if realm is not None and len(realm):
        realm_ptr = <const char*>&realm[0]

    krb5_get_init_creds_opt_set_default_flags(context.raw, appname_ptr, realm_ptr, opt.raw)
